import lepcpy as le
import wx
import wx.lib.newevent
import collections


class LineLabelPanel(wx.Panel):
    WIDTH = 32 * 2
    HEIGHT = 32

    def __init__(self, parent, label, color):
        super().__init__(parent, wx.ID_ANY, size=wx.Size(
            self.WIDTH, self.HEIGHT))
        self.SetForegroundColour('#dbdbdb')
        self.SetBackgroundColour('#000000')
        self.color = color
        self.label = label
        self.Bind(wx.EVT_PAINT, self.__onPaint)

    def __onPaint(self, _):
        self.__drawDc(wx.PaintDC(self))

    def __drawDc(self, dc):
        dc.SetPen(wx.Pen(wx.BLACK, style=wx.TRANSPARENT))
        dc.SetFont(wx.Font(self.HEIGHT // 2, wx.FONTFAMILY_DEFAULT,
                   wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_SEMIBOLD))
        dc.SetBrush(wx.Brush(wx.Colour(self.color), wx.SOLID))
        dc.DrawRectangle(0, 0, self.WIDTH, self.HEIGHT)

        tw, th = dc.GetTextExtent(self.label)
        dc.DrawText(self.label, wx.Point(
            (self.WIDTH - tw) // 2, (self.HEIGHT - th) // 2))

    def paint(self):
        self.__drawDc(wx.ClientDC(self))
        self.Refresh()


class ControlLineRowPanel(wx.Panel):
    MAX_DATALEN = 40
    COL_WIDTH = 32

    WIDTH = COL_WIDTH * MAX_DATALEN
    HEIGHT = 32

    def __init__(self, parent):
        super().__init__(parent, wx.ID_ANY, size=wx.Size(self.WIDTH, self.HEIGHT))
        self.SetForegroundColour('#dbdbdb')
        self.SetBackgroundColour('#000000')
        self.data = collections.deque([], maxlen=self.MAX_DATALEN)
        self.data_count = 0

        # Paint Event
        self.__onDraw()
        self.drawEvent, self.drawEventType = wx.lib.newevent.NewEvent()
        self.Bind(self.drawEventType, self.__onDraw)
        self.Bind(wx.EVT_PAINT, self.__onPaint)

    def __onPaint(self, event):
        wx.BufferedPaintDC(self, self.buffer)

    def __drawDc(self, dc):
        base_height = self.HEIGHT // 4
        # Grid
        grid_height = 3 * base_height
        dc.SetPen(wx.Pen(wx.Colour('#494949'), width=2, style=wx.SHORT_DASH))
        dc.DrawLine(0, grid_height, self.WIDTH, grid_height)

        # Line
        last_bk_pen = wx.Pen(wx.Colour('#000000'), width=16)
        last_fg_pen = wx.Pen(wx.Colour('#dbdbdb'), width=8)
        edge_pen = wx.Pen(wx.Colour('#b6b6b6'), width=2)
        high_pen = wx.Pen(wx.Colour('#db9292'), width=2)
        low_pen = wx.Pen(wx.Colour('#92db92'), width=2)
        high_height = base_height
        low_height = 3 * base_height

        prev_state = False
        for index, state in enumerate(list(self.data)):
            origin_width = index * self.COL_WIDTH
            # Draw state (line)
            if state:
                dc.SetPen(high_pen)
                dc.DrawLine(origin_width, high_height,
                            origin_width + self.COL_WIDTH, high_height)
            else:
                dc.SetPen(low_pen)
                dc.DrawLine(origin_width, low_height,
                            origin_width + self.COL_WIDTH, low_height)

            # Draw state (number)
            state_str = '1' if state else '0'
            sw, sh = dc.GetTextExtent(state_str)
            dc.DrawText(state_str, wx.Point(origin_width +
                        (self.COL_WIDTH - sw) // 2, (self.HEIGHT - sh) // 2))

            # Draw edge
            if index > 0 and prev_state != state:
                dc.SetPen(edge_pen)
                dc.DrawLine(origin_width, low_height,
                            origin_width, high_height)
            prev_state = state

        # Draw last bar
        last_index = self.data_count % self.MAX_DATALEN
        last_x = (last_index * self.COL_WIDTH)
        dc.SetPen(last_bk_pen)
        dc.DrawLine(last_x, 0, last_x, self.HEIGHT)
        dc.SetPen(last_fg_pen)
        dc.DrawLine(last_x, 0, last_x, self.HEIGHT)

    def __onDraw(self, _=None):
        self.buffer = wx.Bitmap(self.WIDTH, self.HEIGHT)
        self.__drawDc(wx.BufferedDC(wx.ClientDC(self), self.buffer))

    def draw(self):
        wx.PostEvent(self, self.drawEvent())

    def append(self, state):
        self.data.append(state)
        self.data_count = self.data_count + 1

    def clear(self):
        self.data.clear()
        self.data_count = 0


class ControlLinePanel(wx.Panel):
    LINES = [
        (le.LE_CONTROLLINE_RS, 'RS'),
        (le.LE_CONTROLLINE_CS, 'CS'),
        (le.LE_CONTROLLINE_DR, 'DR'),
        (le.LE_CONTROLLINE_CD, 'CD'),
        (le.LE_CONTROLLINE_ER, 'ER'),
        (le.LE_CONTROLLINE_SQ, 'SQ'),
        (le.LE_CONTROLLINE_CI, 'CI'),
        (le.LE_CONTROLLINE_EX, 'EX')]

    def __init__(self, parent):
        super().__init__(parent, wx.ID_ANY)
        self.line_dict = {}
        sizer = wx.BoxSizer(wx.VERTICAL)
        for key, label in self.LINES:
            horiz_sizer = wx.BoxSizer(wx.HORIZONTAL)
            horiz_sizer.Add(LineLabelPanel(self, label, '#484848'))
            self.line_dict[key] = ControlLineRowPanel(self)
            horiz_sizer.Add(self.line_dict[key])
            sizer.Add(horiz_sizer)

        self.SetSizerAndFit(sizer)

    def append(self, lines_state):
        for bit, _ in self.LINES:
            self.line_dict[bit].append((lines_state & bit) != bit)

    def paint(self):
        for line in self.line_dict.values():
            line.draw()

    def clear(self):
        for line in self.line_dict.values():
            line.clear()
            line.draw()


class MyApp(wx.Frame):
    ID_BTN_BASE = 100
    ID_BTN_CONNECT_DISCONNECT = ID_BTN_BASE + 1
    ID_BTN_START_STOP = ID_BTN_BASE + 2

    def __init__(self):
        super().__init__(None, wx.ID_ANY, title='Data Line Monitor',
                         style=wx.SYSTEM_MENU | wx.CAPTION | wx.CLOSE_BOX)

        self.is_connected = False
        self.is_started = False

        # Create button panels
        btn_sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.txt_serial = wx.TextCtrl(self, wx.ID_ANY, '')
        self.btn_conn = wx.Button(
            self, self.ID_BTN_CONNECT_DISCONNECT, 'Connect')
        self.btn_start_stop = wx.Button(self, self.ID_BTN_START_STOP, 'Start')
        btn_sizer.AddMany(
            [self.txt_serial, self.btn_conn, self.btn_start_stop])

        # Create line panels
        self.line = ControlLinePanel(self)

        # Layout
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(btn_sizer)
        sizer.Add(self.line)
        self.SetSizerAndFit(sizer)
        self.Show()

        # UI Event
        self.Bind(wx.EVT_CLOSE, self.__onClose)
        self.Bind(wx.EVT_BUTTON, self.__onButton)

    def __onStartStop(self):
        if self.is_started:
            if le.stop():
                self.btn_start_stop.SetLabel('Start')
                self.is_started = False
        else:
            self.line.clear()
            # Start manual simulation
            if le.startManualCallBack(self.__onMonitor, le.LE_MONDATA_FORMAT_2, 0, 0, '\0'):
                self.btn_start_stop.SetLabel('Stop')
                self.is_started = True

    def __onConn(self):
        if self.is_connected:
            if self.is_started:
                self.__onStartStop()
            if le.disconnect():
                self.btn_conn.SetLabel('Connect')
                self.is_connected = False
        else:
            if le.connect(self.txt_serial.GetValue()):
                self.btn_conn.SetLabel('Disconnect')
                self.is_connected = True

    def __onButton(self, event):
        id = event.GetId()
        match id:
            case self.ID_BTN_CONNECT_DISCONNECT:
                self.__onConn()
            case self.ID_BTN_START_STOP:
                self.__onStartStop()

    def __onMonitor(self, frames):
        for frame in frames:
            match frame.framestatus:
                case le.LE_MONSTATUS_SD_DATA_VISIBLE:
                    pass
                case le.LE_MONSTATUS_RD_DATA_VISIBLE:
                    pass
                case le.LE_MONSTATUS_SD_TIMESTAMP:
                    pass
                case le.LE_MONSTATUS_RD_TIMESTAMP:
                    pass
                case le.LE_MONSTATUS_CHANGE_CONTROLLINE:
                    self.line.append(frame.change_controlline.oldstate)
                    self.line.append(frame.change_controlline.newstate)

        self.line.paint()

    def __onClose(self, _):
        le.stop()
        le.disconnect()
        self.Destroy()


app = wx.App()
MyApp()
app.MainLoop()
