// C++
#include <algorithm>
#include <array>
#include <cstdint>
#include <sstream>
#include <string>
#include <tuple>

// pybind11
#include <pybind11/pybind11.h>
#include <pybind11/numpy.h>
#include <pybind11/stl.h>
#include <pybind11/functional.h>

// LEPCLIB (Windows) / libLEPC (Linux) Wrapper
#include "wrapper.h"

#define STRINGIFY(x) #x
#define MACRO_STRINGIFY(x) STRINGIFY(x)

namespace py = pybind11;

PYBIND11_MODULE(bindings, m)
{
  m.doc() = "LEPCLIB_RT2 / libLEPC2 Python bindings";

  /**
   * Setup
   */
  py::class_<LE_SYSTIME>(m, "LE_SYSTIME")
      .def(py::init())
      .def_readwrite("wYear", &LE_SYSTIME::wYear)
      .def_readwrite("wMonth", &LE_SYSTIME::wMonth)
      .def_readwrite("wDayOfWeek", &LE_SYSTIME::wDayOfWeek)
      .def_readwrite("wDay", &LE_SYSTIME::wDay)
      .def_readwrite("wHour", &LE_SYSTIME::wHour)
      .def_readwrite("wMinute", &LE_SYSTIME::wMinute)
      .def_readwrite("wSecond", &LE_SYSTIME::wSecond)
      .def_readwrite("wMilliseconds", &LE_SYSTIME::wMilliseconds)
      .def("__str__", [](LE_SYSTIME &self)
           {
        std::stringstream ss; 
        ss << "wYear(" << self.wYear << "), ";
        ss << "wMonth(" << self.wMonth << "), ";
        ss << "wDayOfWeek(" << self.wDayOfWeek << "), ";
        ss << "wDay(" << self.wDay << "), ";
        ss << "wHour(" << self.wHour << "), ";
        ss << "wMinute(" << self.wMinute << "), ";
        ss << "wSecond(" << self.wSecond << "), ";
        ss << "wMilliseconds(" << self.wMilliseconds << ")";
        return ss.str(); })
      //
      ;

  py::class_<LE_LMWORD>(m, "LE_LMWORD")
      .def(py::init())
      .def_readwrite("wMask", &LE_LMWORD::wMask)
      .def_readwrite("wPattern", &LE_LMWORD::wPattern)
      .def("__str__", [](LE_LMWORD &self)
           {
        std::stringstream ss; 
        ss << "wMask(" << self.wMask << "), ";
        ss << "wPattern(" << self.wPattern << ")";
        return ss.str(); })
      //
      ;

  py::class_<LE_LMBYTE>(m, "LE_LMBYTE")
      .def(py::init())
      .def_readwrite("bMask", &LE_LMBYTE::bMask)
      .def_readwrite("bPattern", &LE_LMBYTE::bPattern)
      .def("__str__", [](LE_LMBYTE &self)
           {
        std::stringstream ss; 
        ss << "bMask(" << self.bMask << "), ";
        ss << "bPattern(" << self.bPattern << ")";
        return ss.str(); })
      //
      ;

  py::class_<LE_LTABLE>(m, "LE_LTABLE")
      .def(py::init())
      .def_readwrite("length", &LE_LTABLE::length)
      .def_property(
          "character",
          [](LE_LTABLE &self)
          {
            return py::array_t<unsigned char>(sizeof(self.character), self.character);
          },
          [](LE_LTABLE &self, const py::array_t<unsigned char> &character)
          {
            if (character.ndim() == 1 && character.size() == sizeof(self.character))
            {
              auto buffer = character.request();
              std::memcpy(self.character, buffer.ptr, sizeof(self.character));
            }
          })
      .def_property(
          "flag",
          [](LE_LTABLE &self)
          {
            return py::array_t<unsigned char>(sizeof(self.flag), self.flag);
          },
          [](LE_LTABLE &self, const py::array_t<unsigned char> &flag)
          {
            if (flag.ndim() == 1 && flag.size() == sizeof(self.flag))
            {
              auto buffer = flag.request();
              std::memcpy(self.flag, buffer.ptr, sizeof(self.flag));
            }
          })
      .def("__str__", [](LE_LTABLE &self)
           {
             std::stringstream ss;
             ss << "length(" << self.length << "), ";
             ss << "character[" << sizeof(self.character) / sizeof(self.character[0]) << "], ";
             ss << "flag[" << sizeof(self.flag) / sizeof(self.flag[0]) << "]";
            return ss.str(); })
      //
      ;

  py::class_<LE_LPROGRAM>(m, "LE_LPROGRAM")
      .def(py::init())
      .def_readwrite("length", &LE_LPROGRAM::length)
      .def_property(
          "program_step",
          [](LE_LPROGRAM &self)
          {
            return py::array_t<unsigned char>({sizeof(self.program_step) / sizeof(self.program_step[0]), sizeof(self.program_step[0]) / sizeof(self.program_step[0][0])}, &self.program_step[0][0]);
          },
          [](LE_LPROGRAM &self, py::array_t<unsigned char> &program_step)
          {
            if (program_step.ndim() == 2)
            {
              auto row = program_step.shape(0);
              auto col = program_step.shape(1);
              if ((sizeof(self.program_step) / sizeof(self.program_step[0])) == row && (sizeof(self.program_step[0]) / sizeof(self.program_step[0][0])) == col)
              {
                auto buffer = program_step.request();
                std::memcpy(self.program_step, buffer.ptr, sizeof(self.program_step));
              }
            }
          })
      .def("__str__", [](LE_LPROGRAM &self)
           {
             std::stringstream ss;
             ss << "length(" << self.length << "), ";
             ss << "program_step[" << sizeof(self.program_step) / sizeof(self.program_step[0]) << "][" << sizeof(self.program_step[0]) / sizeof(self.program_step[0][0]) << "]";
             return ss.str(); })
      //
      ;

  py::class_<LE_TRIGGERFACTOR>(m, "LE_TRIGGERFACTOR")
      .def(py::init())
      .def_readwrite("eFactor", &LE_TRIGGERFACTOR::eFactor)
      .def_readwrite("mfError", &LE_TRIGGERFACTOR::mfError)
      .def_readwrite("fSequence", &LE_TRIGGERFACTOR::fSequence)
      .def_property(
          "mbSd",
          [](LE_TRIGGERFACTOR &self)
          {
            std::array<LE_LMBYTE, sizeof(self.mbSd) / sizeof(self.mbSd[0])> mbSd;
            std::copy(std::begin(self.mbSd), std::end(self.mbSd), std::begin(mbSd));
            return mbSd;
          },
          [](LE_TRIGGERFACTOR &self, const std::array<LE_LMBYTE, sizeof(self.mbSd) / sizeof(self.mbSd[0])> &mbSd)
          {
            std::copy(std::begin(mbSd), std::end(mbSd), std::begin(self.mbSd));
          })
      .def_property(
          "mbRd",
          [](LE_TRIGGERFACTOR &self)
          {
            std::array<LE_LMBYTE, sizeof(self.mbRd) / sizeof(self.mbRd[0])> mbRd;
            std::copy(std::begin(self.mbRd), std::end(self.mbRd), std::begin(mbRd));
            return mbRd;
          },
          [](LE_TRIGGERFACTOR &self, const std::array<LE_LMBYTE, sizeof(self.mbRd) / sizeof(self.mbRd[0])> &mbRd)
          {
            std::copy(std::begin(mbRd), std::end(mbRd), std::begin(self.mbRd));
          })
      .def_property(
          "mbW",
          [](LE_TRIGGERFACTOR &self)
          {
            std::array<LE_LMBYTE, sizeof(self.mbW) / sizeof(self.mbW[0])> mbW;
            std::copy(std::begin(self.mbW), std::end(self.mbW), std::begin(mbW));
            return mbW;
          },
          [](LE_TRIGGERFACTOR &self, const std::array<LE_LMBYTE, sizeof(self.mbW) / sizeof(self.mbW[0])> &mbW)
          {
            std::copy(std::begin(mbW), std::end(mbW), std::begin(self.mbW));
          })
      .def_readwrite("mbLine", &LE_TRIGGERFACTOR::mbLine)
      .def_readwrite("eTimerCounter", &LE_TRIGGERFACTOR::eTimerCounter)
      .def_readwrite("eEdge", &LE_TRIGGERFACTOR::eEdge)
      .def_readwrite("usIdleTime", &LE_TRIGGERFACTOR::usIdleTime)
      //
      ;

  py::class_<LE_TRIGGERACTION>(m, "LE_TRIGGERACTION")
      .def(py::init())
      .def_readwrite("eAction", &LE_TRIGGERACTION::eAction)
      .def_readwrite("bNumber", &LE_TRIGGERACTION::bNumber)
      .def_readwrite("eStopOffset", &LE_TRIGGERACTION::eStopOffset)
      .def_readwrite("usSaveOffsetM", &LE_TRIGGERACTION::usSaveOffsetM)
      .def_readwrite("usSaveOffsetP", &LE_TRIGGERACTION::usSaveOffsetP)
      .def_readwrite("eTimer", &LE_TRIGGERACTION::eTimer)
      .def_readwrite("eTimerAction", &LE_TRIGGERACTION::eTimerAction)
      .def_readwrite("eCounter", &LE_TRIGGERACTION::eCounter)
      .def_readwrite("eCounterAction", &LE_TRIGGERACTION::eCounterAction)
      .def_readwrite("eTriggerNo", &LE_TRIGGERACTION::eTriggerNo)
      .def_readwrite("eTriggerSw", &LE_TRIGGERACTION::eTriggerSw)
      .def_readwrite("eTable", &LE_TRIGGERACTION::eTable)
      .def_readwrite("ulResponseTime", &LE_TRIGGERACTION::ulResponseTime)
      .def_readwrite("eLed", &LE_TRIGGERACTION::eLed)
      .def_readwrite("eLedAct", &LE_TRIGGERACTION::eLedAct)
      //
      ;

  py::class_<LE_DATABERT>(m, "LE_DATABERT")
      .def(py::init())
      .def_readwrite("Savail", &LE_DATABERT::Savail)
      .def_readwrite("R_Bit", &LE_DATABERT::R_Bit)
      .def_readwrite("E_Bit", &LE_DATABERT::E_Bit)
      .def_readwrite("Bit_ER", &LE_DATABERT::Bit_ER)
      .def_readwrite("Loss", &LE_DATABERT::Loss)
      .def_readwrite("R_Blk", &LE_DATABERT::R_Blk)
      .def_readwrite("E_Blk", &LE_DATABERT::E_Blk)
      .def_readwrite("Blk_ER", &LE_DATABERT::Blk_ER)
      .def_readwrite("E_Sec", &LE_DATABERT::E_Sec)
      .def_readwrite("EFS", &LE_DATABERT::EFS)
      //
      ;

  /**
   * Protocol
   */
  using LE_SETUP_PROTOCOL_ASYNC = decltype(LE_SETUP_PROTOCOL::async);
  py::class_<LE_SETUP_PROTOCOL_ASYNC>(m, "LE_SETUP_PROTOCOL_ASYNC")
      .def(py::init())
      .def_readwrite("lSpeedSd", &LE_SETUP_PROTOCOL_ASYNC::lSpeedSd)
      .def_readwrite("lSpeedRd", &LE_SETUP_PROTOCOL_ASYNC::lSpeedRd)
      .def_readwrite("eDataCode", &LE_SETUP_PROTOCOL_ASYNC::eDataCode)
      .def_readwrite("eDataBit", &LE_SETUP_PROTOCOL_ASYNC::eDataBit)
      .def_readwrite("eParity", &LE_SETUP_PROTOCOL_ASYNC::eParity)
      .def_readwrite("eStopBit", &LE_SETUP_PROTOCOL_ASYNC::eStopBit)
      .def_readwrite("eDataFormat", &LE_SETUP_PROTOCOL_ASYNC::eDataFormat)
      .def_readwrite("bFrmEndCharCount", &LE_SETUP_PROTOCOL_ASYNC::bFrmEndCharCount)
      .def_property(
          "cFrmEndChar",
          [](LE_SETUP_PROTOCOL_ASYNC &self)
          {
            return std::tuple<char>(self.cFrmEndChar[0]);
          },
          [](LE_SETUP_PROTOCOL_ASYNC &self, std::tuple<char> cFrmEndChar)
          {
            self.cFrmEndChar[0] = std::get<0>(cFrmEndChar);
          })
      .def_readwrite("cFrmEndTime", &LE_SETUP_PROTOCOL_ASYNC::cFrmEndTime)
      .def_readwrite("eBcc", &LE_SETUP_PROTOCOL_ASYNC::eBcc)
      .def_readwrite("bBeginCharCount", &LE_SETUP_PROTOCOL_ASYNC::bBeginCharCount)
      .def_property(
          "cBeginChar",
          [](LE_SETUP_PROTOCOL_ASYNC &self)
          {
            return std::tuple<char, char>(self.cBeginChar[0], self.cBeginChar[1]);
          },
          [](LE_SETUP_PROTOCOL_ASYNC &self, std::tuple<char, char> cBeginChar)
          {
            self.cBeginChar[0] = std::get<0>(cBeginChar);
            self.cBeginChar[1] = std::get<1>(cBeginChar);
          })
      .def_readwrite("bEndCharCount", &LE_SETUP_PROTOCOL_ASYNC::bEndCharCount)
      .def_property(
          "cEndChar",
          [](LE_SETUP_PROTOCOL_ASYNC &self)
          {
            return std::tuple<char, char>(self.cEndChar[0], self.cEndChar[1]);
          },
          [](LE_SETUP_PROTOCOL_ASYNC &self, std::tuple<char, char> cEndChar)
          {
            self.cEndChar[0] = std::get<0>(cEndChar);
            self.cEndChar[1] = std::get<1>(cEndChar);
          })
      .def_readwrite("bItbCharCount", &LE_SETUP_PROTOCOL_ASYNC::bItbCharCount)
      .def_property(
          "cItbChar",
          [](LE_SETUP_PROTOCOL_ASYNC &self)
          {
            return std::tuple<char>(self.cItbChar[0]);
          },
          [](LE_SETUP_PROTOCOL_ASYNC &self, std::tuple<char> cItbChar)
          {
            self.cItbChar[0] = std::get<0>(cItbChar);
          })
      .def_readwrite("fTransparent", &LE_SETUP_PROTOCOL_ASYNC::fTransparent)
      .def_property(
          "cDleChar",
          [](LE_SETUP_PROTOCOL_ASYNC &self)
          {
            return std::tuple<char>(self.cDleChar[0]);
          },
          [](LE_SETUP_PROTOCOL_ASYNC &self, std::tuple<char> cDleChar)
          {
            self.cDleChar[0] = std::get<0>(cDleChar);
          })
      .def_readwrite("eSequence", &LE_SETUP_PROTOCOL_ASYNC::eSequence)
      //
      ;
  using LE_SETUP_PROTOCOL_SYNC = decltype(LE_SETUP_PROTOCOL::sync);
  py::class_<LE_SETUP_PROTOCOL_SYNC>(m, "LE_SETUP_PROTOCOL_SYNC")
      .def(py::init())
      .def_readwrite("lSpeedSd", &LE_SETUP_PROTOCOL_SYNC::lSpeedSd)
      .def_readwrite("lSpeedRd", &LE_SETUP_PROTOCOL_SYNC::lSpeedRd)
      .def_readwrite("eDataCode", &LE_SETUP_PROTOCOL_SYNC::eDataCode)
      .def_readwrite("eDataBit", &LE_SETUP_PROTOCOL_SYNC::eDataBit)
      .def_readwrite("eParity", &LE_SETUP_PROTOCOL_SYNC::eParity)
      .def_readwrite("eParity", &LE_SETUP_PROTOCOL_SYNC::eClock)
      .def_readwrite("eDataFormat", &LE_SETUP_PROTOCOL_SYNC::eDataFormat)
      .def_readwrite("bSynCharCount", &LE_SETUP_PROTOCOL_SYNC::bSynCharCount)
      .def_property(
          "cSynChar",
          [](LE_SETUP_PROTOCOL_SYNC &self)
          {
            return std::tuple<char, char>(self.cSynChar[0], self.cSynChar[1]);
          },
          [](LE_SETUP_PROTOCOL_SYNC &self, std::tuple<char, char> cSynChar)
          {
            self.cSynChar[0] = std::get<0>(cSynChar);
            self.cSynChar[1] = std::get<1>(cSynChar);
          })
      .def_property(
          "cRstChar",
          [](LE_SETUP_PROTOCOL_SYNC &self)
          {
            return std::tuple<char>(self.cRstChar[0]);
          },
          [](LE_SETUP_PROTOCOL_SYNC &self, std::tuple<char> cRstChar)
          {
            self.cRstChar[0] = std::get<0>(cRstChar);
          })
      .def_readwrite("bRstRepeat", &LE_SETUP_PROTOCOL_SYNC::bRstRepeat)
      .def_readwrite("bSuppressCharCount", &LE_SETUP_PROTOCOL_SYNC::bSuppressCharCount)
      .def_property(
          "cSuppressChar",
          [](LE_SETUP_PROTOCOL_SYNC &self)
          {
            return std::tuple<char>(self.cSuppressChar[0]);
          },
          [](LE_SETUP_PROTOCOL_SYNC &self, std::tuple<char> cSuppressChar)
          {
            self.cSuppressChar[0] = std::get<0>(cSuppressChar);
          })
      .def_readwrite("eBcc", &LE_SETUP_PROTOCOL_SYNC::eBcc)
      .def_readwrite("bBeginCharCount", &LE_SETUP_PROTOCOL_SYNC::bBeginCharCount)
      .def_property(
          "cBeginChar",
          [](LE_SETUP_PROTOCOL_SYNC &self)
          {
            return std::tuple<char, char>(self.cBeginChar[0], self.cBeginChar[1]);
          },
          [](LE_SETUP_PROTOCOL_SYNC &self, std::tuple<char, char> cBeginChar)
          {
            self.cBeginChar[0] = std::get<0>(cBeginChar);
            self.cBeginChar[1] = std::get<1>(cBeginChar);
          })
      .def_readwrite("bEndCharCount", &LE_SETUP_PROTOCOL_SYNC::bEndCharCount)
      .def_property(
          "cEndChar",
          [](LE_SETUP_PROTOCOL_SYNC &self)
          {
            return std::tuple<char, char>(self.cEndChar[0], self.cEndChar[1]);
          },
          [](LE_SETUP_PROTOCOL_SYNC &self, std::tuple<char, char> cEndChar)
          {
            self.cEndChar[0] = std::get<0>(cEndChar);
            self.cEndChar[1] = std::get<1>(cEndChar);
          })
      .def_readwrite("bItbCharCount", &LE_SETUP_PROTOCOL_SYNC::bItbCharCount)
      .def_property(
          "cItbChar",
          [](LE_SETUP_PROTOCOL_SYNC &self)
          {
            return std::tuple<char>(self.cItbChar[0]);
          },
          [](LE_SETUP_PROTOCOL_SYNC &self, std::tuple<char> cItbChar)
          {
            self.cItbChar[0] = std::get<0>(cItbChar);
          })
      .def_readwrite("fTransparent", &LE_SETUP_PROTOCOL_SYNC::fTransparent)
      .def_property(
          "cDleChar",
          [](LE_SETUP_PROTOCOL_SYNC &self)
          {
            return std::tuple<char>(self.cDleChar[0]);
          },
          [](LE_SETUP_PROTOCOL_SYNC &self, std::tuple<char> cDleChar)
          {
            self.cDleChar[0] = std::get<0>(cDleChar);
          })
      .def_readwrite("eSequence", &LE_SETUP_PROTOCOL_SYNC::eSequence)
      //
      ;
  using LE_SETUP_PROTOCOL_HDLC = decltype(LE_SETUP_PROTOCOL::hdlc);
  py::class_<LE_SETUP_PROTOCOL_HDLC>(m, "LE_SETUP_PROTOCOL_HDLC")
      .def(py::init())
      .def_readwrite("lSpeedSd", &LE_SETUP_PROTOCOL_HDLC::lSpeedSd)
      .def_readwrite("lSpeedRd", &LE_SETUP_PROTOCOL_HDLC::lSpeedRd)
      .def_readwrite("eDataCode", &LE_SETUP_PROTOCOL_HDLC::eDataCode)
      .def_readwrite("eClock", &LE_SETUP_PROTOCOL_HDLC::eClock)
      .def_readwrite("eDataFormat", &LE_SETUP_PROTOCOL_HDLC::eDataFormat)
      .def_readwrite("eFcs", &LE_SETUP_PROTOCOL_HDLC::eFcs)
      .def_readwrite("fFilterSd", &LE_SETUP_PROTOCOL_HDLC::fFilterSd)
      .def_property(
          "cFilterSdAddr",
          [](LE_SETUP_PROTOCOL_HDLC &self)
          {
            return std::tuple<char>(self.cFilterSdAddr[0]);
          },
          [](LE_SETUP_PROTOCOL_HDLC &self, std::tuple<char> cFilterSdAddr)
          {
            self.cFilterSdAddr[0] = std::get<0>(cFilterSdAddr);
          })
      .def_readwrite("fFilterRd", &LE_SETUP_PROTOCOL_HDLC::fFilterRd)
      .def_property(
          "cFilterRdAddr",
          [](LE_SETUP_PROTOCOL_HDLC &self)
          {
            return std::tuple<char>(self.cFilterRdAddr[0]);
          },
          [](LE_SETUP_PROTOCOL_HDLC &self, std::tuple<char> cFilterRdAddr)
          {
            self.cFilterRdAddr[0] = std::get<0>(cFilterRdAddr);
          })
      .def_readwrite("eIdleMode", &LE_SETUP_PROTOCOL_HDLC::eIdleMode)
      .def_readwrite("cLeadingCount", &LE_SETUP_PROTOCOL_HDLC::cLeadingCount)
      .def_readwrite("eFrameProtocol", &LE_SETUP_PROTOCOL_HDLC::eFrameProtocol)
      .def_readwrite("ePacketProtocol", &LE_SETUP_PROTOCOL_HDLC::ePacketProtocol)
      .def_property(
          "mbFilterSdEx",
          [](LE_SETUP_PROTOCOL_HDLC &self)
          {
            std::array<LE_LMBYTE, sizeof(self.mbFilterSdEx) / sizeof(self.mbFilterSdEx[0])> mbFilterSdEx;
            std::copy(std::begin(self.mbFilterSdEx), std::end(self.mbFilterSdEx), std::begin(mbFilterSdEx));
            return mbFilterSdEx;
          },
          [](LE_SETUP_PROTOCOL_HDLC &self, const std::array<LE_LMBYTE, sizeof(self.mbFilterSdEx) / sizeof(self.mbFilterSdEx[0])> &mbFilterSdEx)
          {
            std::copy(std::begin(mbFilterSdEx), std::end(mbFilterSdEx), std::begin(self.mbFilterSdEx));
          })
      .def_property(
          "mbFilterRdEx",
          [](LE_SETUP_PROTOCOL_HDLC &self)
          {
            std::array<LE_LMBYTE, sizeof(self.mbFilterRdEx) / sizeof(self.mbFilterRdEx[0])> mbFilterRdEx;
            std::copy(std::begin(self.mbFilterRdEx), std::end(self.mbFilterRdEx), std::begin(mbFilterRdEx));
            return mbFilterRdEx;
          },
          [](LE_SETUP_PROTOCOL_HDLC &self, const std::array<LE_LMBYTE, sizeof(self.mbFilterRdEx) / sizeof(self.mbFilterRdEx[0])> &mbFilterRdEx)
          {
            std::copy(std::begin(mbFilterRdEx), std::end(mbFilterRdEx), std::begin(self.mbFilterRdEx));
          })
      //
      ;
  using LE_SETUP_PROTOCOL_PPP = decltype(LE_SETUP_PROTOCOL::ppp);
  py::class_<LE_SETUP_PROTOCOL_PPP>(m, "LE_SETUP_PROTOCOL_PPP")
      .def(py::init())
      .def_readwrite("lSpeedSd", &LE_SETUP_PROTOCOL_PPP::lSpeedSd)
      .def_readwrite("lSpeedRd", &LE_SETUP_PROTOCOL_PPP::lSpeedRd)
      .def_readwrite("eDataCode", &LE_SETUP_PROTOCOL_PPP::eDataCode)
      .def_readwrite("fSuppress", &LE_SETUP_PROTOCOL_PPP::fSuppress)
      .def_readwrite("eFcs", &LE_SETUP_PROTOCOL_PPP::eFcs)
      //
      ;
  using LE_SETUP_PROTOCOL_IRDA = decltype(LE_SETUP_PROTOCOL::irda);
  py::class_<LE_SETUP_PROTOCOL_IRDA>(m, "LE_SETUP_PROTOCOL_IRDA")
      .def(py::init())
      .def_readwrite("lSpeed", &LE_SETUP_PROTOCOL_IRDA::lSpeed)
      .def_readwrite("eDataCode", &LE_SETUP_PROTOCOL_IRDA::eDataCode)
      .def_readwrite("fSuppress", &LE_SETUP_PROTOCOL_IRDA::fSuppress)
      //
      ;
  using LE_SETUP_PROTOCOL_I2C = decltype(LE_SETUP_PROTOCOL::i2c);
  py::class_<LE_SETUP_PROTOCOL_I2C>(m, "LE_SETUP_PROTOCOL_I2C")
      .def(py::init())
      .def_readwrite("lSpeed", &LE_SETUP_PROTOCOL_I2C::lSpeed)
      .def_readwrite("eDataCode", &LE_SETUP_PROTOCOL_I2C::eDataCode)
      .def_readwrite("eI2cSimMode", &LE_SETUP_PROTOCOL_I2C::eI2cSimMode)
      .def_readwrite("f10BitAddr", &LE_SETUP_PROTOCOL_I2C::f10BitAddr)
      .def_readwrite("sSlaveAddr", &LE_SETUP_PROTOCOL_I2C::sSlaveAddr)
      //
      ;
  using LE_SETUP_PROTOCOL_BURST = decltype(LE_SETUP_PROTOCOL::burst);
  py::class_<LE_SETUP_PROTOCOL_BURST>(m, "LE_SETUP_PROTOCOL_BURST")
      .def(py::init())
      .def_readwrite("eDataCode", &LE_SETUP_PROTOCOL_BURST::eDataCode)
      .def_readwrite("eDataBit", &LE_SETUP_PROTOCOL_BURST::eDataBit)
      .def_readwrite("eParity", &LE_SETUP_PROTOCOL_BURST::eParity)
      .def_readwrite("bFrmEndCharCount", &LE_SETUP_PROTOCOL_BURST::bFrmEndCharCount)
      .def_property(
          "cFrmEndChar",
          [](LE_SETUP_PROTOCOL_BURST &self)
          {
            return std::tuple<char>(self.cFrmEndChar[0]);
          },
          [](LE_SETUP_PROTOCOL_BURST &self, std::tuple<char> cFrmEndChar)
          {
            self.cFrmEndChar[0] = std::get<0>(cFrmEndChar);
          })
      .def_readwrite("cFrmEndTime", &LE_SETUP_PROTOCOL_BURST::cFrmEndTime)
      .def_readwrite("eSequence", &LE_SETUP_PROTOCOL_BURST::eSequence)
      //
      ;
  using LE_SETUP_PROTOCOL_SPI = decltype(LE_SETUP_PROTOCOL::spi);
  py::class_<LE_SETUP_PROTOCOL_SPI>(m, "LE_SETUP_PROTOCOL_SPI")
      .def(py::init())
      .def_readwrite("lSpeed", &LE_SETUP_PROTOCOL_SPI::lSpeed)
      .def_readwrite("eDataCode", &LE_SETUP_PROTOCOL_SPI::eDataCode)
      .def_readwrite("eSpiSimMode", &LE_SETUP_PROTOCOL_SPI::eSpiSimMode)
      .def_readwrite("eSpiCPol", &LE_SETUP_PROTOCOL_SPI::eSpiCPol)
      .def_readwrite("eSpiCPha", &LE_SETUP_PROTOCOL_SPI::eSpiCPha)
      .def_readwrite("lFrmEndTime", &LE_SETUP_PROTOCOL_SPI::lFrmEndTime)
      .def_readwrite("cFrmEndTime", &LE_SETUP_PROTOCOL_SPI::cFrmEndTime)
      .def_readwrite("eSequence", &LE_SETUP_PROTOCOL_SPI::eSequence)
      //
      ;
  using LE_SETUP_PROTOCOL_MODBUS = decltype(LE_SETUP_PROTOCOL::modbus);
  py::class_<LE_SETUP_PROTOCOL_MODBUS>(m, "LE_SETUP_PROTOCOL_MODBUS")
      .def(py::init())
      .def_readwrite("lSpeed", &LE_SETUP_PROTOCOL_MODBUS::lSpeed)
      .def_readwrite("eTransMode", &LE_SETUP_PROTOCOL_MODBUS::eTransMode)
      .def_readwrite("eParity", &LE_SETUP_PROTOCOL_MODBUS::eParity)
      //
      ;
  using LE_SETUP_PROTOCOL_PROFIBUS = decltype(LE_SETUP_PROTOCOL::profibus);
  py::class_<LE_SETUP_PROTOCOL_PROFIBUS>(m, "LE_SETUP_PROTOCOL_PROFIBUS")
      .def(py::init())
      .def_readwrite("lSpeed", &LE_SETUP_PROTOCOL_PROFIBUS::lSpeed)
      //
      ;
  py::class_<LE_SETUP_PROTOCOL>(m, "LE_SETUP_PROTOCOL")
      .def(py::init())
      .def_readwrite("eProtocol", &LE_SETUP_PROTOCOL::eProtocol)
      .def_readwrite("asyn", &LE_SETUP_PROTOCOL::async)
      .def_readwrite("sync", &LE_SETUP_PROTOCOL::sync)
      .def_readwrite("hdlc", &LE_SETUP_PROTOCOL::hdlc)
      .def_readwrite("ppp", &LE_SETUP_PROTOCOL::ppp)
      .def_readwrite("irda", &LE_SETUP_PROTOCOL::irda)
      .def_readwrite("i2c", &LE_SETUP_PROTOCOL::i2c)
      .def_readwrite("burst", &LE_SETUP_PROTOCOL::burst)
      .def_readwrite("spi", &LE_SETUP_PROTOCOL::spi)
      .def_readwrite("modbus", &LE_SETUP_PROTOCOL::modbus)
      .def_readwrite("profibus", &LE_SETUP_PROTOCOL::profibus)
      //
      ;

  /**
   * Monitor data
   */
  py::class_<LE_MONDATAFORMAT1>(m, "LE_MONDATAFORMAT1")
      .def(py::init())
      .def_readwrite("statusSD", &LE_MONDATAFORMAT1::statusSD)
      .def_readwrite("statusRD", &LE_MONDATAFORMAT1::statusRD)
      .def_readwrite("SD", &LE_MONDATAFORMAT1::SD)
      .def_readwrite("RD", &LE_MONDATAFORMAT1::RD)
      //
      ;

  using LE_MONDATAFORMAT2_DATA = decltype(LE_MONDATAFORMAT2::data);
  py::class_<LE_MONDATAFORMAT2_DATA>(m, "LE_MONDATAFORMAT2_DATA")
      .def(py::init())
      .def_readwrite("datastatus", &LE_MONDATAFORMAT2_DATA::datastatus)
      .def_readwrite("SD", &LE_MONDATAFORMAT2_DATA::SD)
      .def_readwrite("RD", &LE_MONDATAFORMAT2_DATA::RD)
      //
      ;
  using LE_MONDATAFORMAT2_TIME = decltype(LE_MONDATAFORMAT2::time);
  py::class_<LE_MONDATAFORMAT2_TIME>(m, "LE_MONDATAFORMAT2_TIME")
      .def(py::init())
      .def_readwrite("h", &LE_MONDATAFORMAT2_TIME::h)
      .def_readwrite("m", &LE_MONDATAFORMAT2_TIME::m)
      .def_readwrite("l", &LE_MONDATAFORMAT2_TIME::l)
      //
      ;
  using LE_MONDATAFORMAT2_CHANGE_CONTROLLINE = decltype(LE_MONDATAFORMAT2::change_controlline);
  py::class_<LE_MONDATAFORMAT2_CHANGE_CONTROLLINE>(m, "LE_MONDATAFORMAT2_CHANGE_CONTROLLINE")
      .def(py::init())
      .def_readwrite("oldstate", &LE_MONDATAFORMAT2_CHANGE_CONTROLLINE::oldstate)
      .def_readwrite("newstate", &LE_MONDATAFORMAT2_CHANGE_CONTROLLINE::newstate)
      //
      ;
  py::class_<LE_MONDATAFORMAT2>(m, "LE_MONDATAFORMAT2")
      .def(py::init())
      .def_readwrite("framestatus", &LE_MONDATAFORMAT2::framestatus)
      .def_readwrite("data", &LE_MONDATAFORMAT2::data)
      .def_readwrite("time", &LE_MONDATAFORMAT2::time)
      .def_readwrite("change_controlline", &LE_MONDATAFORMAT2::change_controlline)
      //
      ;

  py::class_<LE_PROGRESS>(m, "LE_PROGRESS")
      .def(py::init())
      .def_readwrite("progressitem", &LE_PROGRESS::progressitem)
      .def_readwrite("value", &LE_PROGRESS::value)
      //
      ;

  /**
   * Exception
   */
  py::register_exception<wrapper::LEPCError>(m, "LEPCError");

  /**
   * Module definitions
   */
  m.def("getDllVersion", &wrapper::getDllVersion);
  m.def("connect", &wrapper::connect);
  m.def("disconnect", &wrapper::disconnect);
  m.def("getDateTime", &wrapper::getDateTime);
  m.def("setDateTime", &wrapper::setDateTime);
  m.def("startOnlineCallback", &wrapper::startOnlineCallBack);
  m.def("startManualCallBack", &wrapper::startManualCallBack);
  m.def("startBufferCallBack", &wrapper::startBufferCallBack);
  m.def("startProgramCallBack", &wrapper::startProgramCallBack);
  m.def("startProgramCallBack2", &wrapper::startProgramCallBack2);
  m.def("startFlowSendCharCallBack", &wrapper::startFlowSendCharCallBack);
  m.def("startFlowSendLineCallBack", &wrapper::startFlowSendLineCallBack);
  m.def("startFlowRecvCharCallBack", &wrapper::startFlowRecvCharCallBack);
  m.def("startFlowRecvLineCallBack", &wrapper::startFlowRecvLineCallBack);
  m.def("startEchoBufferCallBack", &wrapper::startEchoBufferCallBack);
  m.def("startEchoCharCallBack", &wrapper::startEchoCharCallBack);
  m.def("startEchoLoopbackCallBack", &wrapper::startEchoLoopbackCallBack);
  m.def("startBertContinue", &wrapper::startBertContinue);
  m.def("startBertRbit", &wrapper::startBertRbit);
  m.def("startBertRuntime", &wrapper::startBertRuntime);
  m.def("startBertRepeat", &wrapper::startBertRepeat);
  m.def("restartCallBack", &wrapper::restartCallBack, py::call_guard<py::gil_scoped_release>());
  m.def("startOnline", &wrapper::startOnline);
  m.def("stop", &wrapper::stop, py::call_guard<py::gil_scoped_release>());
  m.def("getRunStatus", &wrapper::getRunStatus);
  m.def("emulationKey", &wrapper::emulationKey);
  m.def("getTable", &wrapper::getTable);
  m.def("setTable", &wrapper::setTable);
  m.def("setTableString", &wrapper::setTableString);
  m.def("getProgram", &wrapper::getProgram);
  m.def("setProgram", &wrapper::setProgram);
  m.def("setTrigger", &wrapper::setTrigger);
  m.def("setTriggerOff", &wrapper::setTriggerOff);
  m.def("getTrigger", &wrapper::getTrigger);
  m.def("getProtocol", &wrapper::getProtocol);
  m.def("setProtocol", &wrapper::setProtocol);
  m.def("getProtocolSetup", &wrapper::getProtocolSetup);
  m.def("setProtocolSetup", &wrapper::setProtocolSetup);

#ifdef __linux__
  m.def("getTimer", &wrapper::getTimer);
  m.def("getCounter", &wrapper::getCounter);
#endif

  m.def("interfaceRS232C", &wrapper::interfaceRS232C);
  m.def("interfaceRS530", &wrapper::interfaceRS530);
  m.def("interfaceTTL", &wrapper::interfaceTTL);
  m.def("interfaceLineControlOff", &wrapper::interfaceLineControlOff);
  m.def("interfaceLineControlDTE", &wrapper::interfaceLineControlDTE);
  m.def("interfaceLineControlDCE", &wrapper::interfaceLineControlDCE);
  m.def("getInterfaceMode", &wrapper::getInterfaceMode);
  m.def("setIdleTimeRec", &wrapper::setIdleTimeRec);
  m.def("getIdleTimeRec", &wrapper::getIdleTimeRec);
  m.def("setTimeStampRec", &wrapper::setTimeStampRec);
  m.def("getTimeStampRec", &wrapper::getTimeStampRec);
  m.def("setLineRec", &wrapper::setLineRec);
  m.def("getLineRec", &wrapper::getLineRec);
  m.def("setCaptureBuf", &wrapper::setCaptureBuf);
  m.def("getCaptureBuf", &wrapper::getCaptureBuf);
  m.def("loadDT", &wrapper::loadDT);
  m.def("isAbort", &wrapper::isAbort);
  m.def("setBscTranslate", &wrapper::setBscTranslate);
  m.def("getBscTranslate", &wrapper::getBscTranslate);
  m.def("getScreen", &wrapper::getScreen);
  m.def("getState", &wrapper::getState);
  m.def("getBertDataCount", &wrapper::getBertDataCount);
  m.def("getBertData", &wrapper::getBertData);
  m.def("importPcap", &wrapper::importPcap);

#ifdef VERSION_INFO
  m.attr("__version__") = MACRO_STRINGIFY(VERSION_INFO);
#else
  m.attr("__version__") = "dev";
#endif
}